package ru.dbotthepony.kstarbound.client.render

import org.apache.logging.log4j.LogManager

const val RenderLayerUpperBits = 5
const val RenderLayerLowerBits = 32 - RenderLayerUpperBits
const val RenderLayerLowerMask = 0.inv() shr RenderLayerUpperBits

enum class RenderLayer(val index: Long) {
	BackgroundOverlay              (1L  shl RenderLayerLowerBits),
	BackgroundTile                 (2L  shl RenderLayerLowerBits),
	Platform                       (3L  shl RenderLayerLowerBits),
	Plant                          (4L  shl RenderLayerLowerBits),
	PlantDrop                      (5L  shl RenderLayerLowerBits),
	Object                         (6L  shl RenderLayerLowerBits),
	PreviewObject                  (7L  shl RenderLayerLowerBits),
	BackParticle                   (8L  shl RenderLayerLowerBits),
	Vehicle                        (9L  shl RenderLayerLowerBits),
	Effect                         (10L shl RenderLayerLowerBits),
	Projectile                     (11L shl RenderLayerLowerBits),
	Monster                        (12L shl RenderLayerLowerBits),
	Npc                            (13L shl RenderLayerLowerBits),
	Player                         (14L shl RenderLayerLowerBits),
	ItemDrop                       (15L shl RenderLayerLowerBits),
	Liquid                         (16L shl RenderLayerLowerBits),
	MiddleParticle                 (17L shl RenderLayerLowerBits),
	ForegroundTile                 (18L shl RenderLayerLowerBits),
	ForegroundEntity               (19L shl RenderLayerLowerBits),
	ForegroundOverlay              (20L shl RenderLayerLowerBits),
	FrontParticle                  (21L shl RenderLayerLowerBits),
	Overlay                        (22L shl RenderLayerLowerBits);

	companion object {
		private val logger = LogManager.getLogger()

		private inline fun perform(value: String, symbol: Char, operator: (Long, Long) -> Long): Long {
			val split = value.split(symbol)

			if (split.size != 2) {
				logger.error("Ambiguous render layer string: $value; assuming 0")
				return 0
			} else {
				val enum = entries.find { it.name == split[0] }

				if (enum == null) {
					logger.error("Unknown render layer: ${split[0]} in $value; assuming 0")
					return 0
				}

				val num = split[1].toLongOrNull()

				if (num == null) {
					logger.error("Invalid render layer string: $value; assuming 0")
					return 0
				}

				return operator(enum.index, num)
			}
		}

		fun parse(value: String): Long {
			if ('+' in value) {
				return perform(value, '+', Long::plus)
			} else if ('-' in value) {
				return perform(value, '-', Long::minus)
			} else if ('*' in value) {
				return perform(value, '*', Long::times)
			} else if ('/' in value) {
				return perform(value, '/', Long::div)
			} else {
				val enum = entries.find { it.name == value }

				if (enum == null) {
					logger.error("Unknown render layer: $value; assuming 0")
					return 0
				}

				return enum.index
			}
		}
	}
}

